import numpy as np

from VLMP.components.modelExtensions import modelExtensionBase

class plates(modelExtensionBase):

    """
    Component name: plates
    Component type: modelExtension

    Author: Pablo Ibáñez-Freire
    Date: 17/06/2023

    Common epsilon, sigma plates for particles in the system.

    :param platesSeparation: Distance between plates.
    :param epsilon: Energy parameter for the plates.
    :param sigma: Length parameter for the plates.
    :param compressionVelocity: Velocity at which the plates are compressed.
    :param minPlatesSeparation: Minimum distance between plates.
    :param maxPlatesSeparation: Maximum distance between plates.

    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {'platesSeparation',
                                                'epsilon','sigma',
                                                'compressionVelocity',
                                                'minPlatesSeparation','maxPlatesSeparation'},
                         requiredParameters  = {"platesSeparation",
                                                "epsilon","sigma"},
                         availableSelections = {"selection"},
                         requiredSelections  = set(),
                         **params)

        ############################################################
        ############################################################
        ############################################################

        platesSeparation = params.get("platesSeparation","auto")

        epsilon         = params["epsilon"]
        sigma           = params["sigma"]

        compressionVelocity = params.get("compressionVelocity",0.0)

        minPlatesSeparation = params.get("minPlatesSeparation",None)
        maxPlatesSeparation = params.get("maxPlatesSeparation",None)

        parameters = {"platesEpsilon":epsilon,
                      "platesSigma":sigma,
                      "compressionVelocity":compressionVelocity}

        if "selection" in params and platesSeparation != "auto":
            self.logger.warning("Selection is ignored when platesSeparation is not set to 'auto'")

        if platesSeparation == "auto":
            # Check if selection is defined
            if "selection" not in params:
                self.logger.error("Selection must be defined when platesSeparation is set to 'auto'")
                raise ValueError("No selection defined")
            else:
                selectedIds  = self.getSelection("selection")
                idsPositions = self.getIdsState(selectedIds,"position")

                if len(idsPositions) == 0:
                    self.logger.error("Selection is empty")
                    raise ValueError("Empty selection")

                platesSeparation = np.max(np.abs(np.array(idsPositions)[:,2]))+sigma
                platesSeparation = 2*platesSeparation

                self.logger.info("Plates separation set to {}".format(platesSeparation))
                parameters["platesSeparation"] = platesSeparation
        else:
            parameters["platesSeparation"] = float(platesSeparation)

        if minPlatesSeparation is not None:
            parameters["minPlatesSeparation"] = minPlatesSeparation

        if maxPlatesSeparation is not None:
            parameters["maxPlatesSeparation"] = maxPlatesSeparation

        extension = {}

        extension[name] = {}
        extension[name]["type"] = ["External","Plates"]
        extension[name]["parameters"] = parameters

        ############################################################

        self.setExtension(extension)



import sys, os

import logging

from . import systemBase

class simulationName(systemBase):
    """
    {"author": "Pablo Ibáñez-Freire",
     "description":
     "Essential component for naming a simulation. This component is compulsory
      in every simulation configuration as each simulation requires a unique name.
      The 'simulationName' component assigns a descriptive and identifiable name
      to a simulation, facilitating its management and reference within the system.
      This name is used as the primary identifier for the simulation across various
      components and modules.",

     "parameters":{
        "simulationName":{"description":"Unique name assigned to the simulation",
                          "type":"str"}
            },

     "example":"
         {
            \"type\":\"simulationName\",
            \"simulationName\":\"MyUniqueSimulationName\"
         }
        "
    }
    """

    availableParameters = {"simulationName"}
    requiredParameters  = {"simulationName"}

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = self.availableParameters,
                         requiredParameters  = self.requiredParameters,
                         **params)

        ############################################################
        ############################################################
        ############################################################

        system = {
            name:{"type":["Simulation","Information"],
                  "parameters":{}}
        }

        system[name]["parameters"]["name"] = params.get("simulationName")

        ############################################################

        self.setSystem(system)




import sys, os

import logging

from . import unitsBase

class none(unitsBase):

    """
    {"author": "Pablo Ibáñez-Freire",
     "description":
     "Component for defining a 'none' unit system in a simulation. This units
      component is used when no specific unit conversions are required.
      It sets all constants, such as the Boltzmann constant
      (KBOLTZ) and the electrostatic coefficient (ELECOEF), to a value of 1. This can
      be particularly useful in simulations where unitless or normalized values are
      preferred, or where specific unit conversions are handled externally.",
     "parameters": {},
     "example":"
         {
            \"type\":\"none\"
         }
        "
    }

    """

    availableParameters = set()
    requiredParameters  = set()

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = self.availableParameters,
                         requiredParameters  = self.requiredParameters,
                         **params)

        ############################################################
        ############################################################
        ############################################################

        self.setUnitsName("None")

        self.addConstant("KBOLTZ",1.0)
        self.addConstant("ELECOEF",1.0)



import sys, os

import copy

import logging

from . import integratorBase

class NVE(integratorBase):
    """
    {"author": "Pablo Ibáñez-Freire",
     "description":
     "NVE integrator for simulations maintaining a constant number of particles,
      volume, and energy. This integrator is suitable for closed-system simulations
      where energy exchange with the environment is not considered.",

     "parameters":{
        "integrationSteps":{"description":"Number of integration steps",
                            "type":"int"},
        "timeStep":{"description":"Time step of the integrator",
                    "type":"float"}
        },

     "example":"
         {
            \"type\":\"NVE\",
            \"integrationSteps\":10000,
            \"timeStep\":0.001
         }
        "
    }
    """

    availableParameters = {"integrationSteps", "timeStep"}
    requiredParameters  = {"integrationSteps", "timeStep"}

    def __init__(self, name, **params):
        super().__init__(_type=self.__class__.__name__, _name=name,
                         availableParameters=self.availableParameters,
                         requiredParameters=self.requiredParameters,
                         **params)

        ############################################################
        ############################################################
        ############################################################

        integrationSteps = params.get("integrationSteps")

        parameters = {}

        parameters["timeStep"] = params.get("timeStep")

        integrator = {
            "type" : ["Verlet","VelocityVerlet"],
            "parameters" : copy.deepcopy(parameters)
        }

        ############################################################

        self.setIntegrationSteps(integrationSteps)
        self.setIntegrator(integrator)



import sys, os
import copy

import logging

import numpy as np

from . import modelExtensionBase

class constraintParticlesListPositionLambda(modelExtensionBase):

    """
    Component name: constraintParticlesListPositionLambda
    Component type: modelExtension

    Author: Pablo Ibáñez-Freire
    Date: 5/12/2023

    Apply a lambda constraint to the position of a set of particles.
    Particles are given by two lists, one with the ids and the other one with the positions

    :param K: Stiffness of the constraint
    :type K: float
    :param n: Exponent of the constraint
    :type n: int
    :param ids: List of particle ids
    :type ids: list of int
    :param positions: List of particle positions
    :type positions: list of list of float

    ...
    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"K","n","ids","positions"},
                         requiredParameters  = {"K","ids","positions"},
                         availableSelections = set(),
                         requiredSelections  = set(),
                         **params)

        ############################################################
        ############################################################
        ############################################################

        K        = params["K"]
        n        = params.get("n",2)

        ids      = params["ids"]
        positions= params["positions"]

        #Check if K is a float
        if not isinstance(K,float) and not isinstance(K,list):
            raise Exception("K must be a float or a list of floats")
        if isinstance(K,float):
            K = [K,K,K]

        #Check if ids and positions are a list
        if not isinstance(ids,list):
            raise Exception("ids must be a list of ints")

        if not isinstance(positions,list):
            raise Exception("positions must be list of floats")

        if len(ids) != len(positions):
            raise Exception("ids and positions must have the same len")

        #Check if ids and positions have the same size
        #if len(ids) != len(positions)

        r0 = [0.0,0.0,0.0]

        extension = {}

        extension[name] = {}
        extension[name]["type"] = ["Bond1","LambdaFixedHarmonicAnisotropic"]
        extension[name]["parameters"] = {"n":n}
        extension[name]["labels"] = ["id_i","position","K","r0"]
        extension[name]["data"]   = []
        for id_,pos in zip(ids,positions):
            extension[name]["data"].append([id_,pos,K,r0])

        ############################################################

        self.setExtension(extension)



import numpy as np

from scipy.optimize import fsolve
from scipy.spatial.transform import Rotation

def helixEquation(s,a,b,e):

    H = np.sqrt(a**2 + b**2)

    x = a*np.cos(s/H)
    y = a*np.sin(s/H)*e
    z = b*s/H

    return np.asarray([x,y,z])

def helixPointsDistanceMinimization(s1,s2,a,b,e,sigma):

    p1 = helixEquation(s1,a,b,e)
    p2 = helixEquation(s2,a,b,e)

    dx = p1[0]-p2[0]
    dy = p1[1]-p2[1]
    dz = p1[2]-p2[2]

    d = np.sqrt(dx*dx+dy*dy+dz*dz)-sigma

    return d

def computeHelixParticlesDistance(a,b,e,sigma):

    try:
        sOne = fsolve(helixPointsDistanceMinimization,sigma,args=(0.0,a,b,e,sigma),xtol=1e-5)[0] # x0 = sigma
        #Check if sOne is correct
        diff = np.abs(helixPointsDistanceMinimization(sOne,0.0,a,b,e,sigma))
        if diff >1e-5:
            print(f"Error computing sOne, diff: {diff}")
            raise Exception("sOne is not correct")
    except:
        print(f"Error computing sOne")
        raise Exception("Error computing sOne")

    return sOne


def discreteHelixFrame(i,sOne,a,b,e):

    pos     = helixEquation( i*sOne,a,b,e)
    posNext = helixEquation((i+1)*sOne,a,b,e)

    dr = posNext-pos
    ex = dr/np.linalg.norm(dr)

    ey = np.cross(np.asarray([0.0,0.0,1.0]),ex)
    ey = ey/np.linalg.norm(ey)

    ez = np.cross(ex,ey)
    ez = ez/np.linalg.norm(ez)

    return np.asarray([ex,ey,ez]).T

def computeHelixMatrix(a,pitch,e,sigma):

    b     = pitch/ (2.0 * np.pi)

    sOne = computeHelixParticlesDistance(a,b,e,sigma)

    R_0 = discreteHelixFrame(0,sOne,a,b,e)
    R_1 = discreteHelixFrame(1,sOne,a,b,e)

    return R_0.T@R_1 # R_1 in the basis of R_0

def computeConnections(a,pitch,helicity,sigma):

    R_H = computeHelixMatrix(a,pitch,helicity,sigma)

    b     = pitch/ (2.0 * np.pi)

    sOne = computeHelixParticlesDistance(a,b,helicity,sigma)

    R_0 = discreteHelixFrame(0,sOne,a,b,helicity)
    R_1 = discreteHelixFrame(1,sOne,a,b,helicity)

    pos0 = helixEquation(0.0,a,b,helicity)
    pos1 = helixEquation(sOne,a,b,helicity)

    pos0 = R_0.T@pos0
    pos1 = R_0.T@pos1

    dr   = (pos1-pos0)
    dr   = dr/np.linalg.norm(dr)
    dr   = dr*sigma/2.0

    connectionNext = dr

    pos0 = helixEquation(0.0,a,b,helicity)
    pos1 = helixEquation(sOne,a,b,helicity)

    pos0 = R_1.T@pos0
    pos1 = R_1.T@pos1

    dr   = (pos0-pos1)
    dr   = dr/np.linalg.norm(dr)
    dr   = dr*sigma/2.0

    connectionPrevious = dr

    return connectionNext,connectionPrevious

def computeLinker(a,pitch,helicity,sigma,angle):

    ez = np.asarray([0.0,0.0,1.0])

    b     = pitch/ (2.0 * np.pi)

    sOne = computeHelixParticlesDistance(a,b,helicity,sigma)

    R    = discreteHelixFrame(0,sOne,a,b,helicity)

    pos0         = helixEquation(0.0,a,b,helicity)
    helixAxisPos = np.asarray([0.0,0.0,0.0])

    dr   = (pos0-helixAxisPos)
    dr   = dr/np.linalg.norm(dr)

    linkerMax = R.T@dr
    linkerMax = linkerMax/np.linalg.norm(linkerMax)
    linkerMax = linkerMax*sigma/2.0

    axis = np.cross(linkerMax,ez)
    axis = axis/np.linalg.norm(axis)

    ## Rotate the dr vector around the axis by angle
    Raxis = Rotation.from_rotvec(axis*angle).as_matrix()

    linker = Raxis@linkerMax
    linker = linker/np.linalg.norm(linker)
    linker = linker*sigma/2.0

    return linker

def generateHelix(N,a,pitch,helicity,sigma,initPos = np.array([0.0,0.0,0.0]),initOri = np.eye(3)):
    """
    Generates a helix with N beads, pitch and helicity
    """

    R_H = computeHelixMatrix(a,pitch,helicity,sigma)

    pos = np.zeros((N,3))
    ori = np.zeros((N,3,3))

    pos[0]     = initPos
    ori[0,:,:] = initOri

    for i in range(1,N):
        pos[i]  = ori[i-1]@np.asarray([sigma,0.0,0.0])+pos[i-1]
        ori[i]  = ori[i-1]@R_H

    return pos,ori



import sys, os

import logging

from . import systemBase

class backup(systemBase):
    """
    {"author": "Pablo Ibáñez-Freire",
     "description":
     " Component used to add a backup system to the simulation.
       When this component is added to the simulation, the simulation will
       create a backup of the simulation every backupIntervalStep steps.
       The simulation will try to restore the simulation from the backup
       if the simulation crashes.",
     "parameters":{
        "backupIntervalStep":{"description":"Interval of steps between backups",
                              "type":"ullint"},
        "backupFilePath":{"description":"Path to the backup file",
                          "type":"str",
                          "default":"backup"},
        "backupStartStep":{"description":"Step at which the backup starts",
                           "type":"ullint",
                           "default":0},
        "backupEndStep":{"description":"Step at which the backup ends",
                          "type":"ullint",
                          "default":"MAX_ULLINT"}
            },
     "example":"
         {
            "type":"backup",
            "backupIntervalStep":1000,
            "backupFilePath":"backup"
         }
        "
    }
    """

    availableParameters  = {"backupIntervalStep",
                            "backupStartStep","backupEndStep",
                            "backupFilePath"}

    requiredParameters = {"backupIntervalStep"}

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = self.availableParameters,
                         requiredParameters  = self.requiredParameters,
                         **params)

        ############################################################
        ############################################################
        ############################################################

        system = {
            name:{"type":["Simulation","Backup"],
                  "parameters":{}}
        }

        system[name]["parameters"]["backupIntervalStep"] = params.get("backupIntervalStep")
        system[name]["parameters"]["backupFilePath"]     = params.get("backupFilePath","backup")

        if "backupStartStep" in params:
            system[name]["parameters"]["backupStartStep"] = params.get("backupStartStep")
        if "backupEndStep" in params:
            system[name]["backupEndStep"] = params.get("backupEndStep")

        ############################################################

        self.setSystem(system)










import sys, os
import copy

import logging

from . import ensembleBase

class NVTlambda(ensembleBase):
    """
    {"author": "Pablo Ibáñez-Freire",
     "description":
     "Component for setting up an NVTlambda ensemble in a simulation. This ensemble type
      extends the standard NVT (constant Number of particles, Volume, and Temperature)
      by introducing an additional lambda parameter, which is essential for thermodynamic
      integration. The component is used to define the simulation environment with a
      fixed box size, temperature, and the lambda value, enabling more complex
      thermodynamic calculations.",

     "parameters":{
        "box":{"description":"Size of the simulation box",
               "type":"float"},
        "temperature":{"description":"Temperature of the simulation environment",
                       "type":"float"},
        "lambda":{"description":"Lambda parameter for thermodynamic integration",
                  "type":"float"}
        },

     "example":"
         {
            \"type\":\"NVTlambda\",
            \"box\":10.0,
            \"temperature\":300.0,
            \"lambda\":0.5
         }
        "
    }
    """

    availableParameters = {"box", "temperature", "lambda"}
    requiredParameters  = {"box", "temperature", "lambda"}

    def __init__(self, name, **params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = self.availableParameters,
                         requiredParameters  = self.requiredParameters,
                         **params)

        ############################################################
        ############################################################
        ############################################################

        box         = params.get("box")
        temperature = params.get("temperature")
        lambda_     = params.get("lambda")

        self.setEnsembleName("NVTlambda")

        self.addEnsembleComponent("box",box)
        self.addEnsembleComponent("temperature",temperature)
        self.addEnsembleComponent("lambda",lambda_)



import sys, os

import copy

import logging

import numpy as np

from . import modelBase

from pyGrained.utils.data import getData

class IDP(modelBase):
    """
    Component name: IDP
    Component type: model

    Author: Pablo Ibáñez-Freire
    Date: 25/03/2023

    Intrinsic disorder protein model.

    Reference: https://journals.aps.org/pre/pdf/10.1103/PhysRevE.90.042709

    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name= name,
                         availableParameters = {"sequence"},
                         requiredParameters  = {"sequence"},
                         definedSelections   = {"particleId"},
                         **params)

        EXCLUSION_DST = 3

        HYDROPHOBICITY_MONERA = {
            'ALA': 0.62,  # Alanine
            'ARG': 0.26,  # Arginine
            'ASN': 0.17,  # Asparagine
            'ASP': 0.0,   # Aspartic Acid
            'CYS': 0.67,  # Cysteine
            'GLN': 0.28,  # Glutamine
            'GLU': 0.15,  # Glutamic Acid
            'GLY': 0.35,  # Glycine
            'HIS': 0.4,   # Histidine
            'ILE': 0.99,  # Isoleucine
            'LEU': 0.98,  # Leucine
            'LYS': 0.2,   # Lysine
            'MET': 0.83,  # Methionine
            'PHE': 1.0,   # Phenylalanine
            'PRO': 0.05,  # Proline
            'SER': 0.32,  # Serine
            'THR': 0.43,  # Threonine
            'TRP': 0.97,  # Tryptophan
            'TYR': 0.75,  # Tyrosine
            'VAL': 0.84   # Valine
        }

        ############################################################


        sequence = params["sequence"]

        kT = self.getEnsemble().getEnsembleComponent("temperature")*self.getUnits().getConstant("KBOLTZ")

        Kb = kT/((0.046)**2)
        r0 = 3.9

        Ka     = kT/((0.26)**2)
        theta0 = 2.12

        self.logger.info(f"kT = {kT}")
        self.logger.info(f"Kb = {Kb}, r0 = {r0}")
        self.logger.info(f"Ka = {Ka}, theta0 = {theta0}")

        sigma       = 4.8
        debyeLength = 9.0
        dielectricConstant = 80.0

        # Energy parameters

        #alphaCG = 0.52
        alphaCG = 0.50

        kes = 1.485 # This is computed for dielectricConstant = 80.0, and sigma = 4.8 !!!

        epsilon_r = 1.0*kT # It does not change.
        epsilon_a = alphaCG*kes*kT

        #

        debyeCutOffFactor = 1.2 # "and the screened Coulomb potential V es is negligible beyond the screening length λ" from reference

        ############################################################
        ######################  Set up model  ######################
        ############################################################

        units = self.getUnits()

        if units.getName() != "KcalMol_A":
            self.logger.error(f"[IDP] Units are not set correctly. Please set units to \"KcalMol_A\" (selected: {units.getName()})")
            raise Exception("Not correct units")

        ############################################################
        # Set up types

        names   = getData("aminoacids")
        masses  = getData("aminoacidMasses")
        radii   = getData("aminoacidRadii")
        charges = getData("aminoacidCharges")

        # In this model the charge of HIS is 0.1
        charges["HIS"] = 0.1

        #Invert names dict
        aminoacids_3to1 = {v: k for k, v in names.items()}

        ############################################################

        types = self.getTypes()
        for nm in names:
            types.addType(name = nm,
                          mass = masses[nm],
                          radius = radii[nm],
                          charge = charges[nm])

        ############################################################

        # State

        state = {}
        state["labels"] = ["id","position"]
        state["data"]   = []

        for i,s in enumerate(sequence):
            pos = [0,0,r0*i]
            # We add small random noise to avoid dihedral singularities
            pos = np.array(pos) + np.random.normal(0,0.01,3)
            pos = pos.tolist()
            state["data"].append([i,pos])

        # Structure

        structure = {}
        structure["labels"] = ["id","type"]
        structure["data"]   = []

        for i,s in enumerate(sequence):
            try:
                structure["data"].append([i,aminoacids_3to1[s]])
            except:
                self.logger.error(f"[IDP] Aminoacid {s} not recognized")
                raise Exception("Aminoacid not recognized")

        # Forcefield

        forcefield = {}

        forcefield["bonds"] = {}
        forcefield["bonds"]["type"]             = ["Bond2","HarmonicCommon_K_r0"]
        forcefield["bonds"]["parameters"]       = {}
        forcefield["bonds"]["parameters"]["K"]  = Kb
        forcefield["bonds"]["parameters"]["r0"] = r0
        forcefield["bonds"]["labels"]           = ["id_i","id_j"]
        forcefield["bonds"]["data"]             = []

        for i in range(len(sequence)-1):
            forcefield["bonds"]["data"].append([i,i+1])

        forcefield["angles"] = {}
        forcefield["angles"]["type"]                 = ["Bond3","HarmonicAngularCommon_K_theta0"]
        forcefield["angles"]["parameters"]           = {}
        forcefield["angles"]["parameters"]["K"]      = Ka
        forcefield["angles"]["parameters"]["theta0"] = theta0
        forcefield["angles"]["labels"]               = ["id_i","id_j","id_k"]
        forcefield["angles"]["data"]                 = []

        for i in range(len(sequence)-2):
            forcefield["angles"]["data"].append([i,i+1,i+2])

        forcefield["dihedrals"] = {}
        forcefield["dihedrals"]["type"]                 = ["Bond4","IDP_Fourier"]
        forcefield["dihedrals"]["parameters"]           = {}
        forcefield["dihedrals"]["labels"]               = ["id_i","id_j","id_k","id_l"]
        forcefield["dihedrals"]["data"]                 = []

        for i in range(len(sequence)-3):
            forcefield["dihedrals"]["data"].append([i,i+1,i+2,i+3])

        # Non bonded

        exclusions = {}
        for i in range(len(sequence)):
            for j in range(i+1,len(sequence)):
                if np.abs(i-j) <= EXCLUSION_DST:
                    exclusions.setdefault(i,[]).append(j)
                    exclusions.setdefault(j,[]).append(i)

        for exc in exclusions:
            exclusions[exc] = sorted(list(set(exclusions[exc])))

        #NL
        forcefield["nl"] = {}
        forcefield["nl"]["type"]       = ["VerletConditionalListSet", "nonExcluded"]
        forcefield["nl"]["parameters"] = {"cutOffVerletFactor": 1.5}
        forcefield["nl"]["labels"]     = ["id", "id_list"]
        forcefield["nl"]["data"] = []

        if EXCLUSION_DST != 0:
            for i in range(len(sequence)):
                forcefield["nl"]["data"].append([i,exclusions[i]])

        # HYDROPHOBIC
        forcefield["hydrophobic"] = {}
        forcefield["hydrophobic"]["type"]       = ["NonBonded", "SplitLennardJones"]
        forcefield["hydrophobic"]["parameters"] = {"cutOffFactor": 2.5,
                                                   "epsilon_r": epsilon_r,
                                                   "epsilon_a": epsilon_a,
                                                   "condition":"nonExcluded"}

        forcefield["hydrophobic"]["labels"] = ["name_i", "name_j", "epsilon", "sigma"]
        forcefield["hydrophobic"]["data"]   = []

        for t1 in names:
            for t2 in names:
                eps = np.sqrt(HYDROPHOBICITY_MONERA[t1]*HYDROPHOBICITY_MONERA[t2])
                forcefield["hydrophobic"]["data"].append([t1,t2,eps,sigma])

        #DH
        forcefield["DH"] = {}
        forcefield["DH"]["type"]       = ["NonBonded", "DebyeHuckel"]
        forcefield["DH"]["parameters"] = {"cutOffFactor": debyeCutOffFactor,
                                          "debyeLength": debyeLength,
                                          "dielectricConstant": dielectricConstant,
                                          "condition":"nonExcluded"}

        ############################################################

        self.setState(state)
        self.setStructure(structure)
        self.setForceField(forcefield)

    def processSelection(self,**params):

        sel = []

        if "particleId" in params:
            sel += params["particleId"]

        return sel



import sys, os

import copy

import logging

from . import integratorBase

class MagneticBrownian(integratorBase):
    """
    {"author": "P. Palacios-Alonso",
     "description":
     "Simple Euler-Maruyama integrator adapted for Brownian dynamics with magnetic
      properties. It is designed for simulating the behavior of magnetically responsive
      particles in a fluid, considering viscosity, gyromagnetic ratio, damping, and
      saturation magnetization.",

     "parameters":{
        "integrationSteps":{"description":"Number of integration steps",
                            "type":"int"},
        "timeStep":{"description":"Time step of the integrator",
                    "type":"float"},
        "viscosity":{"description":"Viscosity of the fluid",
                     "type":"float"},
        "gyroRatio":{"description":"Gyromagnetic ratio",
                     "type":"float"},
        "damping":{"description":"Damping factor",
                   "type":"float"},
        "msat":{"description":"Saturation magnetization",
                "type":"float"},
        "magneticIntegrationAlgorithm":{"description":"Algorithm for magnetic integration",
                                        "type":"string"}
        },

     "example":"
         {
            \"type\":\"MagneticBrownian\",
            \"integrationSteps\":10000,
            \"timeStep\":0.001,
            \"viscosity\":0.01,
            \"gyroRatio\":2.8,
            \"damping\":0.1,
            \"msat\":1.0,
            \"magneticIntegrationAlgorithm\":\"algorithmName\"
         }
        "
    }
    """

    availableParameters = {"integrationSteps", "timeStep", "viscosity", "gyroRatio", "damping", "msat", "magneticIntegrationAlgorithm"}
    requiredParameters  = {"integrationSteps", "timeStep", "viscosity", "gyroRatio", "damping", "msat", "magneticIntegrationAlgorithm"}

    def __init__(self, name, **params):
        super().__init__(_type=self.__class__.__name__, _name=name,
                         availableParameters=self.availableParameters,
                         requiredParameters=self.requiredParameters,
                         **params)

        ############################################################
        ############################################################
        ############################################################

        integrationSteps = params.get("integrationSteps")

        parameters = {}

        parameters["timeStep"]  = params.get("timeStep")
        parameters["viscosity"] = params.get("viscosity")
        parameters["gyroRatio"] = params.get("gyroRatio")
        parameters["damping"]   = params.get("damping")
        parameters["msat"]      = params.get("msat")
        parameters["magneticIntegrationAlgorithm"] = params.get("magneticIntegrationAlgorithm")


        integrator = {
            "type" : ["Magnetic", "Brownian"],
            "parameters" : copy.deepcopy(parameters)
        }

        ############################################################

        self.setIntegrationSteps(integrationSteps)
        self.setIntegrator(integrator)



import sys, os

import logging

from . import typesBase

class none(typesBase):
    """
    {"author": "Pablo Ibáñez-Freire",
     "description":
     "Component representing a 'none' type in a simulation.
      Essentially, it signifies that no additional components or special characteristics are
      associated with the entity. This can be useful in simulations where certain entities need
      to be defined but do not require specific attributes or behaviors.",

     "parameters": {},

     "example":"
         {
            \"type\":\"none\"
         }
        "
    }
    """


    availableParameters = set()
    requiredParameters  = set()

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = self.availableParameters,
                         requiredParameters  = self.requiredParameters,
                         **params)

        ############################################################
        ############################################################
        ############################################################

        self.setTypesName("None")



import sys, os

import logging

from . import simulationStepBase

class anglesMeasurement(simulationStepBase):
    """
    Component name: anglesMeasurement
    Component type: simulationStep

    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"outputFilePath"},
                         requiredParameters  = {"outputFilePath"},
                         availableSelections = {"selection"},
                         requiredSelections  = {"selection"},
                         **params)

        ############################################################
        ############################################################
        ############################################################

        parameters = {}

        parameters["outputFilePath"] = params["outputFilePath"]

        selIds = self.getSelection("selection")

        #Check if selIds is a list of list of size 3
        if isinstance(selIds,list):
            for selId in selIds:
                if isinstance(selId,list):
                    if len(selId) != 3:
                        self.logger.error("[anglesMeasurement] selection must be a list of list of size 3")
                        raise Exception("Selection error")
                else:
                    self.logger.error("[anglesMeasurement] selection must be a list of list of size 3")
                    raise Exception("Selection error")
        else:
            self.logger.error("[anglesMeasurement] selection must be a list of list of size 3")
            raise Exception("Selection error")

        simulationStep = {
            name:{
              "type":["ParticlesListMeasure","AnglesMeasure"],
              "parameters":{**parameters},
              "labels":["id_i","id_j","id_k"],
              "data":selIds.copy()
            }
        }

        ############################################################

        self.setSimulationStep(simulationStep)






import sys, os

import logging

from . import simulationStepBase

class lambdaCycle(simulationStepBase):
    """
    Component name: lambdaCycle
    Component type: simulationStep

    Author: Pablo Ibáñez-Freire
    Date: 03/01/2023

    Lambda cycle.

    :param activationStep: Activation step.
    :type activationStep: int
    :param measureStep: Measure step.
    :type measureStep: int
    :param pauseStep: Pause step.
    :type pauseStep: int
    :param lambdaValues: Lambda values.
    :type lambdaValues: list

    """

    def __init__(self,name,**params):
        # Chech if interval step is defined
        if 'intervalStep' in params:
            # Check if interval step is equal to 1
            if params['intervalStep'] != 1:
                # Raise error
                raise ValueError('Interval step must be equal to 1 for lambdaCycle component')
        else:
            # Set interval step to 1
            params['intervalStep'] = 1
        super().__init__(_type= self.__class__.__name__,
                         _name= name,
                         availableParameters = {"activationStep","measureStep","pauseStep","lambdaValues"},
                         requiredParameters  = {"activationStep","measureStep","pauseStep","lambdaValues"},
                         availableSelections = set(),
                         requiredSelections  = set(),
                         **params)

        ############################################################
        ############################################################
        ############################################################

        activationStep = params['activationStep']
        measureStep    = params['measureStep']
        pauseStep      = params['pauseStep']

        lambdaValues   = params['lambdaValues']

        #Check if lambdaValues is a list and all elements are numbers between 0 and 1
        if not isinstance(lambdaValues,list):
            self.logger.error('lambdaValues must be a list')
            raise TypeError('lambdaValues must be a list')
        else:
            for i in lambdaValues:
                if not isinstance(i,(int,float)):
                    self.logger.error('lambdaValues must be a list of numbers')
                    raise TypeError('lambdaValues must be a list of numbers')
                elif i < 0 or i > 1:
                    self.logger.error('lambdaValues must be a list of numbers between 0 and 1')
                    raise ValueError('lambdaValues must be a list of numbers between 0 and 1')

        # The first element of lambdaValues must be 0 and the last one must be 1
        if lambdaValues[0] != 0:
            self.logger.error('The first element of lambdaValues must be 0')
            raise ValueError('The first element of lambdaValues must be 0')
        elif lambdaValues[-1] != 1:
            self.logger.error('The last element of lambdaValues must be 1')
            raise ValueError('The last element of lambdaValues must be 1')

        simulationStep = {
            name:{
                  "type":["FlowControl","LambdaCycle"],
                  "parameters":{"activationStep":activationStep,
                                "measureStep":measureStep,
                                "pauseStep":pauseStep,
                                "lambdaValues":lambdaValues}
            }
        }

        ############################################################

        self.setSimulationStep(simulationStep)





from VLMP.components.modelOperations import modelOperationBase

import numpy as np

class setParticleXYPosition(modelOperationBase):

    """
    Component name: setParticleXYPosition
    Component type: modelOperation

    Author: Pablo Ibáñez-Freire
    Date: 26/02/2024

    Set the XY particle position to value.

    :param position: Position to set the XY particle to.
    :type position: float list
    :param considerRadius: Consider particle radius when setting the XY position.
    :type considerRadius: bool, optional

    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"position"},
                         requiredParameters  = {"position"},
                         availableSelections = {"selection"},
                         requiredSelections  = {"selection"},
                         **params)

        ############################################################
        ############################################################
        ############################################################

        # Check if the given position is a list of two floats
        if not isinstance(params["position"],list) or len(params["position"]) != 2:
            self.logger.error("The given position must be a list of two floats.")
            raise Exception("Parameter bad format")

        targetPosition = params["position"]

        selectedIds = self.getSelection("selection")
        pos = np.asarray(self.getIdsState(selectedIds,"position"))

        for i in range(len(selectedIds)):
            pos[i][0] = targetPosition[0]
            pos[i][1] = targetPosition[1]

        self.setIdsState(selectedIds,"position",pos.tolist())



import sys, os

import logging

from . import simulationStepBase

class forceBetweenSetsMeasurement(simulationStepBase):
    """
    Component name: forceBetweenSetsMeasurement
    Component type: simulationStep

    Author: Pablo Palacios-Alonso
    Date: 12/02/2024

    """


    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"setName_idList",
                                                "outputFilePath"},
                         requiredParameters  = {"setName_idList",
                                                "outputFilePath"},
                         availableSelections = set(),
                         requiredSelections  = set(),
                         **params)

        ############################################################
        ############################################################
        ############################################################
        
        parameters = {}
        
        parameters["outputFilePath"] = params["outputFilePath"]

        data = []
        setName_idList = params["setName_idList"]
        
        for setName in setName_idList.keys():
            data += [[setName, setName_idList[setName]]]
        
        
        simulationStep = {
            name:{
                "type":["MechanicalMeasure","ForceBetweenSetsMeasure"],
                "parameters":{**parameters},
                "labels":["name", "id_list"],
                "data":data.copy()
            }
        }

        ############################################################

        self.setSimulationStep(simulationStep)






import sys, os

import logging

from . import systemBase

class seed(systemBase):
    """
    {"author": "Pablo Ibáñez-Freire",
     "description":
     "This component sets the seed for the random number generator",
     "parameters":{
        "seed":{"description":"Seed for the random number generator",
                "type":"int"}
            },

     "example":"
         {
            \"type\":\"seed\",
            \"seed\":123456
         }
        "
    }
    """

    availableParameters = {"seed"}
    requiredParameters  = {"seed"}

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = self.availableParameters,
                         requiredParameters  = self.requiredParameters,
                         **params)

        ############################################################
        ############################################################
        ############################################################

        system = {
            name:{"type":["Simulation","Information"],
                  "parameters":{}}
        }

        system[name]["parameters"]["seed"] = params.get("seed")

        ############################################################

        self.setSystem(system)




import sys, os

import json

import copy

import logging

import numpy as np
from scipy.spatial.transform import Rotation as R

from scipy.spatial import KDTree

from ..types import basic

from . import modelBase

from . import IDP # intrinsic disorder protein model
from . import ENM # protein model

import pyUAMMD

class PROTEIN_IDP_PROTEIN(modelBase):
    """
    Component name: PROTEIN_IDP_PROTEIN
    Component type: model

    Author: Pablo Ibáñez-Freire
    Date: 02/01/2024

    Intrinsic disorder protein model merged with protein model (ENM)

    Reference: https://journals.aps.org/pre/pdf/10.1103/PhysRevE.90.042709

    """
    def __loadProtein(self,name,PDB,types,PRTN_types):

        typesPDB = basic(name="PDB_types",units= self.getUnits())

        prot = ENM.ENM(name = name + "_PRTN",
                       units = self.getUnits(),
                       types = typesPDB,
                       ensemble = self.getEnsemble(),
                       PDB = PDB)

        state      = prot.getState()
        structure  = prot.getStructure()
        forcefield = prot.getForceField()

        structure_str = json.dumps(structure)

        # Rename all keys in forcefield adding the suffix "_PRTN"
        forcefield = {key + "_PRTN":value for key,value in forcefield.items()}
        forcefield_str = json.dumps(forcefield)

        # Rename all types in typesPDB adding the suffix "_PRTN"
        for _,t in typesPDB.getTypes().items():
            newName = t["name"] + "_PRTN"
            structure_str = structure_str.replace(t["name"],newName)
            forcefield_str = forcefield_str.replace(t["name"],newName)
            t["name"] = newName
            PRTN_types.add(newName)
            types.addType(**t)

        structure = json.loads(structure_str)
        forcefield = json.loads(forcefield_str)

        # Iterate over forcefield each non-bonded interaction, now acts on typeGroup2
        for key,value in forcefield.items():
            entryType,_ = value["type"]
            if entryType == "NonBonded":
                forcefield[key]["parameters"]["condition"] = "typeGroup2Intra"

        protSim = pyUAMMD.simulation({"state":state, "topology":{ "structure":structure, "forceField":forcefield}})

        return copy.deepcopy(protSim)

    def __addConnection(self,forceField,ids_conn_start,ids_conn_end):

        # 0-1 2-3

        # bonds: 2-3
        bonds = [[ids_conn_start[1],ids_conn_end[0]]]

        # angles: 0-1-2 1-2-3
        angles = [[ids_conn_start[0],ids_conn_start[1],ids_conn_end[0]],
                  [ids_conn_start[1],ids_conn_end[0],ids_conn_end[1]]]

        # dihedrals: 0-1-2-3
        dihedrals = [[ids_conn_start[0],ids_conn_start[1],ids_conn_end[0],ids_conn_end[1]]]

        exclusions = {}
        for i in ids_conn_start + ids_conn_end:
            exclusions[i] = []

        exclusions[bonds[0][0]].extend([bonds[0][1]])
        exclusions[bonds[0][1]].extend([bonds[0][0]])

        exclusions[angles[0][0]].extend([angles[0][1],angles[0][2]])
        exclusions[angles[0][1]].extend([angles[0][0],angles[0][2]])
        exclusions[angles[0][2]].extend([angles[0][0],angles[0][1]])

        exclusions[angles[1][0]].extend([angles[1][1],angles[1][2]])
        exclusions[angles[1][1]].extend([angles[1][0],angles[1][2]])
        exclusions[angles[1][2]].extend([angles[1][0],angles[1][1]])

        #exclusions[dihedrals[0][0]].extend([dihedrals[0][1],dihedrals[0][2],dihedrals[0][3]])
        #exclusions[dihedrals[0][1]].extend([dihedrals[0][0],dihedrals[0][2],dihedrals[0][3]])
        #exclusions[dihedrals[0][2]].extend([dihedrals[0][0],dihedrals[0][1],dihedrals[0][3]])
        #exclusions[dihedrals[0][3]].extend([dihedrals[0][0],dihedrals[0][1],dihedrals[0][2]])

        ##############################################################

        addBonds     = False
        addAngles    = False
        addDihedrals = False
        for entry in forceField:
            if forceField[entry]["type"][0] == "Bond2":
                if "IDP" in entry:
                    forceField[entry]["data"].extend(bonds)
                    addBonds = True
            elif forceField[entry]["type"][0] == "Bond3":
                if "IDP" in entry:
                    forceField[entry]["data"].extend(angles)
                    addAngles = True
            elif forceField[entry]["type"][0] == "Bond4":
                if "IDP" in entry:
                    forceField[entry]["data"].extend(dihedrals)
                    addDihedrals = True

        if not addBonds:
            self.logger.error("No Bond2 IDP entry found in forcefield")
            raise RuntimeError("Entry not found")
        if not addAngles:
            self.logger.error("No Bond3 IDP entry found in forcefield")
            raise RuntimeError("Entry not found")
        if not addDihedrals:
            self.logger.error("No Bond4 IDP entry found in forcefield")
            raise RuntimeError("Entry not found")

        nlIdIndex     = forceField["nl"]["labels"].index("id")
        nlIdListIndex = forceField["nl"]["labels"].index("id_list")

        for k in exclusions:
            found = False
            for excl in forceField["nl"]["data"]:
                if k == excl[nlIdIndex]:
                    excl[nlIdListIndex].extend(exclusions[k])
                    excl[nlIdListIndex] = list(set(excl[nlIdListIndex]))
                    found = True
                    break
            if not found:
                forceField["nl"]["data"].append([k,exclusions[k]])

    def __fixProteinPos(self,sim,IDPids,PDBids,modeIDP,modePDB,maxTries):

        positionsIndex = sim["state"]["labels"].index("position")

        IDPpositions = []
        for i in IDPids:
            IDPpositions.append(sim["state"]["data"][i][positionsIndex])

        PDBpositions = []
        for i in PDBids:
            PDBpositions.append(sim["state"]["data"][i][positionsIndex])

        dstIDP = np.array(IDPpositions[1]) - np.array(IDPpositions[0])
        dstIDP = np.linalg.norm(dstIDP)

        if modeIDP == "start":
            idp_sel_pos = np.array(IDPpositions[0])  + np.asarray([0,0,-dstIDP])
        elif modeIDP == "end":
            idp_sel_pos = np.array(IDPpositions[-1]) + np.asarray([0,0, dstIDP])
        else:
            self.logger.error("Mode not recognized")
            raise RuntimeError("Mode not implemented")

        if modePDB == "start":
            pdb_sel_pos = np.array(PDBpositions[0])
            ignoreIndex = 0
        elif modePDB == "end":
            pdb_sel_pos = np.array(PDBpositions[-1])
            ignoreIndex = len(PDBpositions) - 1
        else:
            self.logger.error("Mode not recognized")
            raise RuntimeError("Mode not implemented")

        translation = pdb_sel_pos - idp_sel_pos

        # Translate protein positions
        for i in range(len(PDBpositions)):
            PDBpositions[i] = np.array(PDBpositions[i]) - translation

        center = idp_sel_pos

        # We rotate randomly the protein, the center of rotation is the selected position
        # Then we compute the minimal distance between protein and IDP.
        # We use KDTree

        tree = KDTree(IDPpositions)

        cTry    = 0
        overlap = True
        self.logger.debug("Fixing protein position ...")

        while overlap:

            if cTry >= maxTries:
                self.logger.error("Max tries reached")
                raise RuntimeError("Max tries reached")

            if cTry == 0:
                pass
            else:
                rotation = R.random()

                for i in range(len(PDBpositions)):
                    PDBpositions[i] = rotation.apply(PDBpositions[i] - center) + center

            overlap = False
            for i, point in enumerate(PDBpositions):
                distance, index = tree.query(point)
                # Check if the distance is within the threshold
                if distance < dstIDP*1.01 and i != ignoreIndex:
                    overlap = True
                    break

            cTry += 1
            self.logger.debug(f"Try {cTry}/{maxTries} ...")

        for i,index in enumerate(PDBids):
            sim["state"]["data"][index][positionsIndex] = PDBpositions[i].tolist()


    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name= name,
                         availableParameters = {"sequence",
                                                "PDB1","PDB2",
                                                "PDB1_conn","PDB2_conn",
                                                "cutOffVerletFactor"},
                         requiredParameters  = {"sequence"},
                         definedSelections   = {"particleId"},
                         **params)

        maxTries = 1000

        seq  = params["sequence"]

        PDB1 = params.get("PDB1",None)
        PDB2 = params.get("PDB2",None)

        cutOffVerletFactor = params.get("cutOffVerletFactor",1.1)

        if PDB1 is not None:
            PDB1_conn = params.get("PDB1_conn","end")
            if PDB1_conn not in ["end","start"]:
                self.logger.error("PDB1 connection must be either 'end' or 'start'")
                raise Exception("Error in input parameters")

        if PDB2 is not None:
            PDB2_conn = params.get("PDB2_conn","end")
            if PDB2_conn not in ["end","start"]:
                self.logger.error("PDB2 connection must be either 'end' or 'start'")
                raise Exception("Error in input parameters")

        if "PDB1_conn" in params and "PDB1" not in params:
            self.logger.error("PDB1 connection defined but not PDB1")
            raise Exception("Error in input parameters")

        if "PDB2_conn" in params and "PDB2" not in params:
            self.logger.error("PDB2 connection defined but not PDB2")
            raise Exception("Error in input parameters")

        types = self.getTypes()

        IDP_types  = set()
        PRTN_types = set()

        # IDP

        typesIDP = basic(name="IDP_types",units= self.getUnits())

        idp = IDP(name = name + "_IDP",
                  units = self.getUnits(),
                  types = typesIDP,
                  ensemble = self.getEnsemble(),
                  sequence = seq)

        state      = idp.getState()
        structure  = idp.getStructure()
        forcefield = idp.getForceField()

        structure_str = json.dumps(structure)

        # Rename all keys in forcefield adding the suffix "_IDP"
        forcefield = {key + "_IDP":value for key,value in forcefield.items()}
        forcefield_str = json.dumps(forcefield)

        # Rename all types in typesIDP adding the suffix "_IDP"
        for _,t in typesIDP.getTypes().items():
            newName = t["name"] + "_IDP"
            structure_str  = structure_str.replace(t["name"],newName)
            forcefield_str = forcefield_str.replace(t["name"],newName)
            t["name"] = newName
            IDP_types.add(newName)
            types.addType(**t)

        structure  = json.loads(structure_str)
        forcefield = json.loads(forcefield_str)

        # Iterate over forcefield each non-bonded interaction, now acts on typeGroup1
        for key,value in forcefield.items():
            entryType,_ = value["type"]
            if entryType == "NonBonded":
                forcefield[key]["parameters"]["condition"] = "typeGroup1Intra"

        sim = pyUAMMD.simulation({"state":state, "topology":{ "structure":structure, "forceField":forcefield}})

        IDP_len = sim.getNumberOfParticles()

        ############################################################

        # PROT1
        if PDB1 is not None:
            prot1Sim = self.__loadProtein(name,PDB1,types,PRTN_types)

        # PROT2
        if PDB2 is not None:
            prot2Sim = self.__loadProtein(name,PDB2,types,PRTN_types)

        ############################################################

        # Merge three simulations

        if PDB1 is not None:
            sim.append(prot1Sim,mode = "modelId")
        if PDB2 is not None:
            sim.append(prot2Sim,mode = "modelId")

        forceField = copy.deepcopy(sim["topology"]["forceField"])

        # Remove all neighbor lists
        labels = ["id","id_list"]
        data   = []

        entriesToRemove = []
        for entry in forceField:
            entryType,_ = forceField[entry]["type"]
            if entryType == "VerletConditionalListSet":

                cutOffVerletFactor = min(cutOffVerletFactor,forceField[entry]["parameters"]["cutOffVerletFactor"])
                data += forceField[entry]["data"]

                entriesToRemove.append(entry)

        for entry in entriesToRemove:
            del forceField[entry]

        forceField["nl"] = {}
        forceField["nl"]["type"]       = ["VerletConditionalListSet","nonExclTypeGroup1Intra_nonExclTypeGroup2Intra_nonExclInter_nonExclNoGroup"]
        forceField["nl"]["parameters"] = {"cutOffVerletFactor":cutOffVerletFactor,
                                          "typeGroup1":sorted(list(IDP_types)),
                                          "typeGroup2":sorted(list(PRTN_types))}

        forceField["nl"]["labels"]     = labels
        forceField["nl"]["data"]       = data

        ############################################################
        # Add connections between IDP and PROT

        IDP_start = [0,1]
        IDP_end   = [IDP_len-2,IDP_len-1]

        # <Prot1> ------- <IDP> ------- <Prot2>

        IDPids    = [i for i in range(IDP_len)]

        if PDB1 is not None:

            PDB1len = prot1Sim.getNumberOfParticles()-1 # Since id starts at 0
            if PDB1_conn == "start":
                ids     = [IDP_len+1,IDP_len]
            else:
                # PDB1_conn == "end"
                ids     = [IDP_len+PDB1len-1,IDP_len+PDB1len]

            self.__addConnection(forceField,ids,IDP_start)
            PDBids = [i for i in range(IDP_len, IDP_len+PDB1len+1)]
            self.__fixProteinPos(sim,IDPids,PDBids,"start",PDB1_conn,maxTries)

            PDB1offset = IDP_len+PDB1len+1
        else:
            PDB1offset = IDP_len

        if PDB2 is not None:

            PDB2len = prot2Sim.getNumberOfParticles()-1
            if PDB2_conn == "start":
                ids     = [PDB1offset,PDB1offset+1]
            else:
                # PDB2_conn == "end"
                ids     = [PDB1offset+PDB2len,PDB1offset+PDB2len-1]

            self.__addConnection(forceField,IDP_end,ids)
            PDBids = [i for i in range(PDB1offset, PDB1offset+PDB2len+1)]
            self.__fixProteinPos(sim,IDPids,PDBids,"end",PDB2_conn,maxTries)

        ############################################################

        self.setState(copy.deepcopy(sim["state"]))
        self.setStructure(copy.deepcopy(sim["topology"]["structure"]))
        self.setForceField(copy.deepcopy(forceField))

    def processSelection(self,**params):

        sel = []

        if "particleId" in params:
            sel += params["particleId"]

        return sel



import sys, os

import copy

import logging

from . import integratorBase

class BBK(integratorBase):
    """
    {"author": "Pablo Ibáñez-Freire",
     "description":
     "BBK integrator for the NVT ensemble. This Langevin integrator is designed for
      simulations that maintain a constant number of particles, volume, and temperature.
      It is particularly useful for molecular dynamics simulations requiring stochastic
      thermal noise and friction.",

     "parameters":{
        "integrationSteps":{"description":"Number of integration steps",
                            "type":"int"},
        "timeStep":{"description":"Time step of the integrator",
                    "type":"float"},
        "frictionConstant":{"description":"Friction constant of the integrator",
                            "type":"float"}
        },

     "example":"
         {
            \"type\":\"BBK\",
            \"integrationSteps\":10000,
            \"timeStep\":0.001,
            \"frictionConstant\":0.1
         }
        "
    }
    """

    availableParameters = {"integrationSteps", "timeStep", "frictionConstant"}
    requiredParameters  = {"integrationSteps", "timeStep", "frictionConstant"}

    def __init__(self, name, **params):
        super().__init__(_type=self.__class__.__name__, _name=name,
                         availableParameters=self.availableParameters,
                         requiredParameters=self.requiredParameters,
                         **params)

        ############################################################
        ############################################################
        ############################################################

        integrationSteps = params.get("integrationSteps")

        parameters = {}

        parameters["timeStep"]         = params.get("timeStep")
        parameters["frictionConstant"] = params.get("frictionConstant")

        integrator = {
            "type" : ["Langevin","BBK"],
            "parameters" : copy.deepcopy(parameters)
        }

        ############################################################

        self.setIntegrationSteps(integrationSteps)
        self.setIntegrator(integrator)



import sys, os
import copy

import logging

from . import ensembleBase

class NVT(ensembleBase):
    """
    {"author": "Pablo Ibáñez-Freire",
     "description":
     "Component for setting up an NVT (constant Number of particles, Volume, and Temperature)
      ensemble in a simulation. This component is used to define the simulation environment
      with a fixed box size and temperature.
      It is essential for simulations that require a controlled temperature and volume,
      commonly used in molecular dynamics and other statistical mechanics simulations.",

     "parameters":{
        "box":{"description":"Size of the simulation box",
               "type":"float"},
        "temperature":{"description":"Temperature of the simulation environment",
                       "type":"float"}
        },

     "example":"
         {
            \"type\":\"NVT\",
            \"box\":10.0,
            \"temperature\":300.0
         }
        "
    }
    """

    availableParameters = {"box", "temperature"}
    requiredParameters  = {"box", "temperature"}

    def __init__(self, name, **params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = self.availableParameters,
                         requiredParameters  = self.requiredParameters,
                         **params)

        ############################################################
        ############################################################
        ############################################################

        box         = params.get("box")
        temperature = params.get("temperature")

        self.setEnsembleName("NVT")

        self.addEnsembleComponent("box",box)
        self.addEnsembleComponent("temperature",temperature)



import sys, os

import logging

import numpy as np

from VLMP.components.models import modelBase

class STERIC_LAMBDA_SOLVATION(modelBase):

    """
    Component name: STERIC_LAMBDA_SOLVATION
    Component type: model

    Author: Pablo Ibáñez-Freire
    Date: 02/01/2024

    steric lambda solvation model.

    :param concentration: Concentration of the solute in the solvent (in N/V units)
    :type concentration: float
    :param condition: Condition for the interaction. Options: "inter", "intra" ...
    :type condition: str, default="inter"
    :param epsilon: epsilon parameter of the steric potential
    :type epsilon: float
    :param cutOffFactor: Factor to multiply the sigma parameter to obtain the cut-off distance.
    :type cutOffFactor: float
    :alpha: alpha parameter of the steric potential
    :type alpha: float, default=0.5
    :param addVerletList: If True, a Verlet list will be created for the interactions.
    :type addVerletList: bool, optional, default=True
    :param particleName: Name of the particle to be added to the system.
    :type particleName: str, optional, default="W"
    :param particleMass: Mass of the particle to be added to the system.
    :type particleMass: float, optional, default=1.0
    :param particleRadius: Radius of the particle to be added to the system.
    :type particleRadius: float, optional, default=1.0
    :param particleCharge: Charge of the particle to be added to the system.
    :type particleCharge: float, optional, default=0.0
    :param padding: Padding to be added to the box to place the particle.
    :type padding: two lists of three floats, optional, default=[[0.0,0.0,0.0],[0.0,0.0,0.0]]

    ...
    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"concentration",
                                                "epsilon","cutOffFactor",
                                                "alpha",
                                                "addVerletList","condition",
                                                "particleName",
                                                "particleMass","particleRadius","particleCharge",
                                                "padding"},
                         requiredParameters  = {"concentration",},
                         definedSelections   = {"particleId"},
                         **params)

        ############################################################
        ############################################################
        ############################################################

        concentration = params["concentration"]

        cutOffFactor  = params.get("cutOffFactor",1.5)
        epsilon       = params.get("epsilon",1.0)
        alpha         = params.get("alpha",0.5)

        addVerletList = params.get("addVerletList",True)
        condition     = params.get("condition","inter")

        particleName  = params.get("particleName","W")

        particleMass   = params.get("particleMass",1.0)
        particleRadius = params.get("particleRadius",1.0)
        particleCharge = params.get("particleCharge",0.0)

        padding = params.get("padding",[[0.0,0.0,0.0],[0.0,0.0,0.0]])

        ############################################################
        # Adding particles

        box = self.getEnsemble().getEnsembleComponent("box")

        # Applying padding
        xMax = +box[0]/2.0 + padding[0][0]
        xMin = -box[0]/2.0 - padding[1][0]

        yMax = +box[1]/2.0 + padding[0][1]
        yMin = -box[1]/2.0 - padding[1][1]

        zMax = +box[2]/2.0 + padding[0][2]
        zMin = -box[2]/2.0 - padding[1][2]

        V   = (xMax-xMin)*(yMax-yMin)*(zMax-zMin)
        N   = int(concentration*V)

        self.logger.info(f"Adding {N} particles of type {particleName} to the system.")
        # Note that the number of paricles is computed after the padding is applied

        # Add type
        types = self.getTypes()
        types.addType(name = particleName,
                      mass = particleMass,
                      radius = particleRadius,
                      charge = particleCharge)

        state = {}
        state["labels"] = ["id","position"]
        state["data"]   = []
        for i in range(N):
            self.logger.debug(f"Adding particle {i+1}/{N} ...")
            x = np.random.uniform(xMin,xMax)
            y = np.random.uniform(yMin,yMax)
            z = np.random.uniform(zMin,zMax)
            state["data"].append([i,[x,y,z]])

        structure = {}
        structure["labels"] = ["id","type"]
        structure["data"]   = []
        for i in range(N):
            structure["data"].append([i,particleName])

        ############################################################

        forceField = {}

        forcefield = {}

        if addVerletList:
            forcefield["nl"]={}
            forcefield["nl"]["type"]       =  ["VerletConditionalListSet",  "nonExclIntra_nonExclInter"]
            forcefield["nl"]["parameters"] =  {}
            forcefield["nl"]["labels"]      = ["id","id_list"]
            forcefield["nl"]["data"]        = []

        forcefield[name] = {}
        forcefield[name]["type"] = ["NonBonded","Steric12SoftCore"]
        forcefield[name]["parameters"] = {"cutOffFactor":cutOffFactor,
                                          "condition":condition,
                                          "alpha":alpha}
        forcefield[name]["labels"] = ["name_i","name_j","epsilon","sigma"]
        forcefield[name]["data"]   = []

        types = self.getTypes().getTypes()
        for t1 in types:
            for t2 in types:
                r1 = types[t1]["radius"]
                r2 = types[t2]["radius"]
                sigma = r1+r2
                forcefield[name]["data"].append([t1,t2,epsilon,sigma])

        ############################################################

        self.setState(state)
        self.setStructure(structure)
        self.setForceField(forcefield)

    def processSelection(self,**params):

        sel = []

        if "particleId" in params:
            sel += [0]

        return sel







import sys, os

import logging

from . import unitsBase

class KcalMol_A(unitsBase):
    """
    {"author": "Pablo Ibáñez-Freire",
     "description":
     "Component for defining the unit system of (Kcal/mol)/A in a simulation. This unit
      component, part of the 'units' category, specifies the system's energy and distance
      units as kilocalories per mole and Angstroms, respectively. It includes fundamental
      constants relevant to this unit system, such as the Boltzmann constant (KBOLTZ) and
      the electrostatics coefficient (ELECOEF).",
     "parameters": {},
     "example":"
         {
            \"type\":\"KcalMol_A\"
         }
        "
    }
    """

    availableParameters = set()
    requiredParameters  = set()

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = self.availableParameters,
                         requiredParameters  = self.requiredParameters,
                         **params)

        ############################################################
        ############################################################
        ############################################################

        self.setUnitsName("KcalMol_A")

        self.addConstant("KBOLTZ",1.987191E-03)
        self.addConstant("ELECOEF",332.0716)



import sys, os
import itertools

import logging

from . import modelBase
from ...utils.geometry import getEx
from ...utils.geometry import boxChecker,platesChecker

from ...utils.input import getSubParameters

from ...utils.geometry.objects.helix import computeHelixMatrix, generateHelix, computeConnections, computeLinker

import random

import orthopoly

import numpy as np
from scipy.spatial import cKDTree

from pyquaternion import Quaternion

from scipy.spatial.transform import Rotation
from scipy.optimize import root


class HELIX(modelBase):
    """
    Component name: HELIX
    Component type: model

    Author: Pablo Ibáñez-Freire
    Date: 25/04/2023

    Polimerization model for a helix.

    """

    def __checkVariantParameters(self,variantRequiredParameters,variantAvailableParameters):
        #Check parameters
        for vPar in variantRequiredParameters:
            if vPar not in self.variantParams:
                self.logger.error(f"[HELIX] Requiered parameter {vPar} for variant {self.variantName} not found."
                                  f"Required parameters are {variantRequiredParameters}")
                raise Exception(f"Required parameter not given")

        #Check parameters
        for vPar in self.variantParams:
            if vPar not in variantAvailableParameters:
                self.logger.error(f"[HELIX] Parameter {vPar} for variant {self.variantName} not found."
                                  f"Available parameters are {variantAvailableParameters}")
                raise Exception(f"Parameter not available")

    def __generateRandomPositions(self):
        monomersPositions    = []
        monomersOrientations = []

        boxX,boxY,boxZ = [b/2.0 for b in self.box]

        n=1
        while(n<=self.nMonomers):
            self.logger.debug(f"[HELIX] Trying to add monomer {n}")

            #We take into account the monomer radius for avoiding problems with PBC
            x=np.random.uniform(low=-boxX + self.monomerRadius, high=boxX - self.monomerRadius)
            y=np.random.uniform(low=-boxY + self.monomerRadius, high=boxY - self.monomerRadius)
            z=np.random.uniform(low=-boxZ + self.monomerRadius, high=boxZ - self.monomerRadius)

            currentMonomerPosition = np.asarray([x,y,z])

            if not self.checker.check([x,y,z]):
                continue

            if monomersPositions:
                minDst,minDstIndex = cKDTree(monomersPositions).query(currentMonomerPosition, 1)
            else:
                minDst = np.inf

            if minDst > 1.5*(2.0*self.monomerRadius):
                monomersPositions.append(currentMonomerPosition)
                q=Quaternion.random()
                q0,q1,q2,q3 = q
                monomersOrientations.append(np.asarray([q0,q1,q2,q3]))
                self.logger.debug(f"[HELIX] Added monomer {n}")
                n=n+1


        return np.asarray(monomersPositions),np.asarray(monomersOrientations)

    def __generateLine(self):
        monomersPositions    = []
        monomersOrientations = []

        boxX,boxY,boxZ = [b/2.0 for b in self.box]

        for n in range(self.nMonomers):
            self.logger.debug(f"[HELIX] Trying to add monomer {n}")

            x=-boxX+self.monomerRadius+(n+1)*2.2*self.monomerRadius
            y=0.0
            z=0.0

            currentMonomerPosition = np.asarray([x,y,z])

            monomersPositions.append(currentMonomerPosition)
            monomersOrientations.append(np.asarray([1.0,0.0,0.0,0.0]))

        centroid = np.mean(monomersPositions,axis=0)
        monomersPositions-=centroid

        for p in monomersPositions:
            if not self.checker.check(p):
                self.logger.error(f"[HELIX] Box too small")
                raise Exception("Box too small")


        return np.asarray(monomersPositions),np.asarray(monomersOrientations)

    def __generateHelix(self):

        monomersPositions,monomersOrientations = generateHelix(self.nMonomers,
                                                               self.helixRadius,self.helixPitch,
                                                               self.helicity,self.monomerRadius*2.0)

        monomersOrientationsTmp = []
        for o in monomersOrientations:
            q = Rotation.from_matrix(o).as_quat() # scalar last
            q = [q[3],q[0],q[1],q[2]] # scalar first
            monomersOrientationsTmp.append(q)

        monomersOrientations = monomersOrientationsTmp.copy()

        centroid = np.mean(monomersPositions,axis=0)
        monomersPositions-=centroid

        for p in monomersPositions:
            if not self.checker.check(p):
                self.logger.error(f"[HELIX] Box too small")
                raise Exception("Box too small")

        return np.asarray(monomersPositions),np.asarray(monomersOrientations)

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name= name,
                         availableParameters = {"mode","init",
                                                "nMonomers",
                                                "monomerRadius",
                                                "epsilon_mm",
                                                "helixRadius","helixPitch",
                                                "helicity",
                                                "variant"},
                         requiredParameters  = {"nMonomers",
                                                "epsilon_mm",
                                                "helixRadius","helixPitch",
                                                "variant"},
                         definedSelections   = {"particleId"},
                         **params)

        self.logger = logging.getLogger("VLMP")

        ############################################################
        ####################  Model Parameters  ####################
        ############################################################

        self.init                 = params.get("init","random")
        self.mode,self.modeParams = getSubParameters("mode",params)

        if self.mode not in ["bulk","surface"]:
            self.logger.error(f"[HELIX] Mode {self.mode} not supported")
            raise Exception(f"Mode not supported")

        self.nMonomers      = params["nMonomers"]

        self.kT  = self.getEnsemble().getEnsembleComponent("temperature")*self.getUnits().getConstant("KBOLTZ")
        self.box = self.getEnsemble().getEnsembleComponent("box")

        self.monomerRadius = params.get("monomerRadius",0.5)

        self.epsilon_mm = params["epsilon_mm"]

        self.helixRadius = params["helixRadius"]
        self.helixPitch  = params["helixPitch"]

        self.helicity = params.get("helicity",1.0)

        ##############################################

        self.R_H = computeHelixMatrix(self.helixRadius,self.helixPitch,self.helicity,
                                      self.monomerRadius*2.0)

        self.e_x = list(self.R_H[:,0])
        self.e_y = list(self.R_H[:,1])
        self.e_z = list(self.R_H[:,2])

        connectionNext,connectionPrevious = computeConnections(self.helixRadius,self.helixPitch,self.helicity,
                                                               self.monomerRadius*2.0)

        self.e_next = list(connectionNext)
        self.e_prev = list(connectionPrevious)

        ##############################################

        # Available variants:
        #   - fixedCosine
        #   - fixedExponential
        #   - dynamicCosine
        #   - dynamicExponential
        #   - twoStatesCosine
        #   - twoStatesExponential

        self.variantName, self.variantParams = getSubParameters("variant",params)

        if self.variantName == "fixedCosine":

            self.logger.info(f"[HELIX] Using fixedCosine variant")

            variantAvailableParameters  = {"E","Kb","theta_start","theta_end","phi_start","phi_end"}
            variantRequiredParameters   = {"E","Kb","theta_start","theta_end","phi_start","phi_end"}

            self.__checkVariantParameters(variantRequiredParameters,variantAvailableParameters)

            if self.init != "helix" and self.init != "line":
                self.logger.error(f"[HELIX] Init mode {self.init} is not avaible for fixedCosine variant")
                raise Exception("Init mode not available")

        elif self.variantName == "fixedExponential":

            self.logger.info(f"[HELIX] Using fixedExponential variant")

            variantAvailableParameters  = {"E","Kb","Ka","Kd"}
            variantRequiredParameters   = {"E","Kb","Ka","Kd"}

            self.__checkVariantParameters(variantRequiredParameters,variantAvailableParameters)

            #Chekc init mode is helix
            if self.init != "helix" and self.init != "line":
                self.logger.error(f"[HELIX] Init mode {self.init} is not avaible for fixedExponential variant")
                raise Exception("Init mode not available")

        elif self.variantName == "dynamicCosine":

            self.logger.info(f"[HELIX] Using dynamicCosine variant")

            variantAvailableParameters  = {"energyThreshold","Eb","theta_start","theta_end","phi_start","phi_end","r_start","rc","patchRadius"}
            variantRequiredParameters   = {"Eb","theta_start","theta_end","phi_start","phi_end","r_start","rc"}

            self.__checkVariantParameters(variantRequiredParameters,variantAvailableParameters)

            if "energyThreshold" not in self.variantParams:
                self.variantParams["energyThreshold"] = 0.0

            if "patchRadius" not in self.variantParams:
                self.variantParams["patchRadius"] = self.monomerRadius/5.0

        elif self.variantName == "dynamicExponential":

            self.logger.info(f"[HELIX] Using dynamicExponential variant")

            variantAvailableParameters  = {"energyThreshold","Eb","Ka","Kd","Kb","rc","patchRadius"}
            variantRequiredParameters   = {"Eb","Ka","Kd","Kb","rc"}

            self.__checkVariantParameters(variantRequiredParameters,variantAvailableParameters)

            if "energyThreshold" not in self.variantParams:
                self.variantParams["energyThreshold"] = 0.0

            if "patchRadius" not in self.variantParams:
                self.variantParams["patchRadius"] = self.monomerRadius/5.0

        elif self.variantName == "twoStatesCosine":

            self.logger.info(f"[HELIX] Using twoStatesCosine variant")

            variantAvailableParameters  = {"energyThreshold","patchRadius",
                                           "theta_start_0","theta_end_0","phi_start_0","phi_end_0",
                                           "theta_start_1","theta_end_1","phi_start_1","phi_end_1",
                                           "r_start_0","r_start_1","rc_0","rc_1",
                                           "prob_0_to_1","prob_1_to_0"}

            variantRequiredParameters   = {"patchRadius",
                                           "theta_start_0","theta_end_0","phi_start_0","phi_end_0",
                                           "theta_start_1","theta_end_1","phi_start_1","phi_end_1",
                                           "r_start_0","r_start_1","rc_0","rc_1",
                                           "prob_0_to_1","prob_1_to_0"}

            self.__checkVariantParameters(variantRequiredParameters,variantAvailableParameters)

            if "energyThreshold" not in self.variantParams:
                self.variantParams["energyThreshold"] = 0.0

            if "patchRadius" not in self.variantParams:
                self.variantParams["patchRadius"] = self.monomerRadius/5.0

        elif self.variantName == "twoStatesExponential":

            self.logger.info(f"[HELIX] Using twoStatesExponential variant")

            variantAvailableParameters  = {"energyThreshold","patchRadius",
                                           "Ka_0","Kd_0","Kb_0",
                                           "Ka_1","Kd_1","Kb_1",
                                           "Kb_0","Kb_1",
                                           "rc_0","rc_1",
                                           "prob_0_to_1","prob_1_to_0"}

            variantRequiredParameters   = {"patchRadius",
                                           "Ka_0","Kd_0","Kb_0",
                                           "Ka_1","Kd_1","Kb_1",
                                           "Kb_0","Kb_1",
                                           "rc_0","rc_1",
                                           "prob_0_to_1","prob_1_to_0"}

            self.__checkVariantParameters(variantRequiredParameters,variantAvailableParameters)

            if "energyThreshold" not in self.variantParams:
                self.variantParams["energyThreshold"] = 0.0

            if "patchRadius" not in self.variantParams:
                self.variantParams["patchRadius"] = self.monomerRadius/5.0

        else:
            self.logger.error(f"[HELIX] Variant {self.variantName} not available")
            raise Exception("Variant not available")

        ##############################################

        if self.mode=="surface":
            #Check parameters Es,beta0,El,Sl are present in params
            requiredSurfaceParams = ["Es","beta0","El","Sl","plateTop","plateBottom"]

            for sPar in requiredSurfaceParams:
                if sPar not in self.modeParams:
                    self.logger.error(f"[HELIX] Requiered parameter {sPar} for surface mode not found."
                                      f"Required parameters are {requiredSurfaceParams}")
                    raise Exception(f"Required parameter not given")

            plateTop    = self.modeParams["plateTop"]
            plateBottom = self.modeParams["plateBottom"]

            self.checker = platesChecker(self.box,plateTop,plateBottom,2.0*self.monomerRadius)
        else:
            self.checker = boxChecker(self.box)

        ##############################################

        self.logger.info(f"[HELIX] Generating HELIX model with {self.nMonomers} monomers")

        if   self.init == "random":
            self.monomersPositions,self.monomersOrientations = self.__generateRandomPositions()
        elif self.init == "line":
            self.monomersPositions,self.monomersOrientations = self.__generateLine()
        elif self.init == "helix":
            self.monomersPositions,self.monomersOrientations = self.__generateHelix()
        else:
            self.logger.error(f"[HELIX] Init mode {self.init} is not avaible")
            raise Exception("Init mode not available")

        ##############################################

        types = self.getTypes()

        types.addType(name="M",radius=self.monomerRadius)

        ##############################################

        state = {}
        state["labels"]=["id","position","direction"]
        state["data"]  =[]

        for i,[p,q] in enumerate(zip(self.monomersPositions,self.monomersOrientations)):
            state["data"].append([int(i),list(p),list(q)])

        ##############################################

        structure = {}
        structure["labels"] = ["id", "type", "modelId"]
        structure["data"]   = []

        for i in range(self.nMonomers):
            structure["data"].append([int(i),"M",int(i)])

        ##############################################

        forceField = {}

        #Verlet list
        forceField["verletListParticles"]={}
        forceField["verletListParticles"]["type"]       =  ["VerletConditionalListSet", "all"]
        forceField["verletListParticles"]["parameters"] =  {"cutOffVerletFactor":1.5}

        #WCA
        forceField["MonomerMonomer"]={}
        forceField["MonomerMonomer"]["type"]       =  ["NonBonded", "GeneralLennardJonesType2"]
        forceField["MonomerMonomer"]["parameters"] =  {"cutOffFactor":2.5,"condition":"all"}
        forceField["MonomerMonomer"]["labels"]     =  ["name_i", "name_j", "epsilon", "sigma"]
        forceField["MonomerMonomer"]["data"]       =  [["M", "M", self.epsilon_mm, 2.0*self.monomerRadius]]

        ############Helix

        if "fixed" in self.variantName:

            forceField["helix"]={}
            if self.variantName == "fixedCosine":
                forceField["helix"]["type"]       = ["Bond2", "HelixCosine"]
            elif self.variantName == "fixedExponential":
                forceField["helix"]["type"]       = ["Bond2", "HelixExponential"]

            forceField["helix"]["parameters"] =  self.variantParams.copy()

            forceField["helix"]["parameters"]["e_x"] = self.e_x
            forceField["helix"]["parameters"]["e_y"] = self.e_y
            forceField["helix"]["parameters"]["e_z"] = self.e_z

            forceField["helix"]["parameters"]["e_next"] = self.e_next
            forceField["helix"]["parameters"]["e_prev"] = self.e_prev

            forceField["helix"]["labels"] = ["id_i","id_j"]
            forceField["helix"]["data"]   = []

            for i in range(self.nMonomers-1):
                forceField["helix"]["data"].append([i,i+1])

        if "dynamic" in self.variantName or "twoStates" in self.variantName:

            patchRadius = self.variantParams["patchRadius"]

            #Helix
            forceField["helix"]={}
            forceField["helix"]["type"]       = ["PatchyParticles","DynamicallyBondedPatchyParticles"]

            forceField["helix"]["patchesState"]={}
            forceField["helix"]["patchesState"]["labels"]=["id","position"]
            forceField["helix"]["patchesState"]["data"]  =[]

            for i in range(self.nMonomers):
                index=i*2
                p = self.e_prev
                forceField["helix"]["patchesState"]["data"].append([int(index  ),list(p)])
                p = self.e_next
                forceField["helix"]["patchesState"]["data"].append([int(index+1),list(p)])

            forceField["helix"]["patchesGlobal"]={}
            forceField["helix"]["patchesGlobal"]["fundamental"] = {}
            forceField["helix"]["patchesGlobal"]["fundamental"]["type"]       = ["Fundamental","DynamicallyBondedPatchyParticles"]
            forceField["helix"]["patchesGlobal"]["fundamental"]["parameters"] = {"energyThreshold":self.variantParams["energyThreshold"]}

            forceField["helix"]["patchesGlobal"]["types"]  = {}
            forceField["helix"]["patchesGlobal"]["types"]["type"]   = ["Types","Basic"]
            forceField["helix"]["patchesGlobal"]["types"]["labels"] = ["name", "mass", "radius", "charge"]
            forceField["helix"]["patchesGlobal"]["types"]["data"]   = [["S",0.0,patchRadius,0.0],
                                                                       ["E",0.0,patchRadius,0.0]]

            forceField["helix"]["patchesTopology"]={}

            forceField["helix"]["patchesTopology"]["structure"]={}
            forceField["helix"]["patchesTopology"]["structure"]["labels"] = ["id", "type", "parentId"]
            forceField["helix"]["patchesTopology"]["structure"]["data"]   = []

            for i in range(self.nMonomers):
                index=i*2
                forceField["helix"]["patchesTopology"]["structure"]["data"].append([index  ,"E",i])
                forceField["helix"]["patchesTopology"]["structure"]["data"].append([index+1,"S",i])

            forceField["helix"]["patchesTopology"]["forceField"] = {}

            #Verlet list
            forceField["helix"]["patchesTopology"]["forceField"]["verletList"]={}
            forceField["helix"]["patchesTopology"]["forceField"]["verletList"]["type"]       =  ["VerletConditionalListSet", "interDifferentType"]
            forceField["helix"]["patchesTopology"]["forceField"]["verletList"]["parameters"] =  {"cutOffVerletFactor":3.0}

            forceField["helix"]["patchesTopology"]["forceField"]["helix"]={}
            forceField["helix"]["patchesTopology"]["forceField"]["helix"]["parameters"] =  {"condition":"inter"}

            if self.variantName == "dynamicCosine":
                forceField["helix"]["patchesTopology"]["forceField"]["helix"]["type"]       =  ["NonBondedPatches", "HelixCosine"]
                forceField["helix"]["patchesTopology"]["forceField"]["helix"]["labels"]     =  ["name_i", "name_j", "Eb", "theta_start", "theta_end", "phi_start", "phi_end", "r_start", "rc", "e_x", "e_y", "e_z"]
                forceField["helix"]["patchesTopology"]["forceField"]["helix"]["data"]       =  [["S", "E",
                                                                                                 self.variantParams["Eb"],
                                                                                                 self.variantParams["theta_start"], self.variantParams["theta_end"],
                                                                                                 self.variantParams["phi_start"], self.variantParams["phi_end"],
                                                                                                 self.variantParams["r_start"], self.variantParams["rc"],
                                                                                                 self.e_x,self.e_y,self.e_z]]
            elif self.variantName == "dynamicExponential":
                forceField["helix"]["patchesTopology"]["forceField"]["helix"]["type"]       =  ["NonBondedPatches", "HelixExponential"]
                forceField["helix"]["patchesTopology"]["forceField"]["helix"]["labels"]     =  ["name_i", "name_j", "Eb", "Kb", "Ka", "Kd", "rc", "e_x", "e_y", "e_z"]
                forceField["helix"]["patchesTopology"]["forceField"]["helix"]["data"]       =  [["S", "E",
                                                                                                 self.variantParams["Eb"], self.variantParams["Kb"],
                                                                                                 self.variantParams["Ka"], self.variantParams["Kd"],
                                                                                                 self.variantParams["rc"],
                                                                                                 self.e_x,self.e_y,self.e_z]]

            else:
                #Not implemented yet
                raise Exception("Not implemented yet")

        #############Patches

        #Surface
        if(self.mode == "surface"):

            beta0  = self.modeParams["beta0"]
            linker = computeLinker(self.helixRadius,self.helixPitch,self.helicity,2.0*self.monomerRadius,beta0)

            if "patchRadius" not in self.variantParams:
                patchRadius = self.monomerRadius/5.0

            forceField["plateUp"]={}
            forceField["plateUp"]["type"]           = ["Surface", "SurfaceWCAType2"]
            forceField["plateUp"]["parameters"]     = {"surfacePosition": plateTop}
            forceField["plateUp"]["labels"]         = ["name", "epsilon", "sigma"]
            forceField["plateUp"]["data"]           = [["M",1.0,2.0*self.monomerRadius]]

            forceField["plateBottom"]={}
            forceField["plateBottom"]["type"]       = ["Surface", "ParabolaSurface"]
            forceField["plateBottom"]["parameters"] = {"surfacePosition":plateBottom+self.monomerRadius}
            forceField["plateBottom"]["labels"]     = ["name", "epsilon"]
            forceField["plateBottom"]["data"]       = [["M",self.modeParams["Es"]]]

            #Patches

            forceField["surface"]={}
            forceField["surface"]["type"]       = ["PatchyParticles","PatchyParticles"]

            forceField["surface"]["patchesState"]={}
            forceField["surface"]["patchesState"]["labels"]=["id","position"]
            forceField["surface"]["patchesState"]["data"]  =[]

            for i in range(self.nMonomers):
                index=i
                forceField["surface"]["patchesState"]["data"].append([int(index),list(linker)])

            forceField["surface"]["patchesGlobal"]={}
            forceField["surface"]["patchesGlobal"]["fundamental"] = {}
            forceField["surface"]["patchesGlobal"]["fundamental"]["type"]      = ["Fundamental","None"]
            forceField["surface"]["patchesGlobal"]["fundamental"]["parameters"]= {}

            forceField["surface"]["patchesGlobal"]["types"]  = {}
            forceField["surface"]["patchesGlobal"]["types"]["type"]   = ["Types","Basic"]
            forceField["surface"]["patchesGlobal"]["types"]["labels"] = ["name", "mass", "radius", "charge"]
            forceField["surface"]["patchesGlobal"]["types"]["data"]   = [["L",0.0,patchRadius,0.0]]

            forceField["surface"]["patchesTopology"] = {}

            forceField["surface"]["patchesTopology"]["structure"] = {}
            forceField["surface"]["patchesTopology"]["structure"]["labels"] = ["id", "type", "parentId"]
            forceField["surface"]["patchesTopology"]["structure"]["data"]   = []

            for i in range(self.nMonomers):
                forceField["surface"]["patchesTopology"]["structure"]["data"].append([i,"L",i])

            forceField["surface"]["patchesTopology"]["forceField"] = {}

            forceField["surface"]["patchesTopology"]["forceField"]["linker"]={}
            forceField["surface"]["patchesTopology"]["forceField"]["linker"]["type"]       =  ["SurfacePatches", "Linker"]
            forceField["surface"]["patchesTopology"]["forceField"]["linker"]["parameters"] =  {"surfacePosition":plateBottom}
            forceField["surface"]["patchesTopology"]["forceField"]["linker"]["labels"]     =  ["name", "epsilon", "sigma"]
            forceField["surface"]["patchesTopology"]["forceField"]["linker"]["data"]       =  [["L",
                                                                                                self.modeParams["El"],
                                                                                                self.modeParams["Sl"]]]

        ##############################################

        self.setState(state)
        self.setStructure(structure)
        self.setForceField(forceField)

    def processSelection(self,**params):

        sel = []
        if "particleId" in params:
            sel += params["particleId"]
        return sel



import sys, os

import logging

from . import typesBase

class basic(typesBase):
    """
    {"author": "Pablo Ibáñez-Freire",
     "description":
     "Component for defining basic types in a simulation.
      It includes components for mass, radius, and charge, which are essential attributes
      for many simulation scenarios. These properties define the physical characteristics
      of the particles used in the simulation. When a new type is created, if no value is
      given for any of these properties, the default value is used.",
     "parameters":{
        "mass":{"description":"Default mass for a new type",
                "type":"float",
                "default":1.0},
        "radius":{"description":"Default radius for a new type",
                  "type":"float",
                  "default":1.0},
        "charge":{"description":"Default charge for a new type",
                  "type":"float",
                  "default":0.0}
            },
     "example":"
         {
            \"type\":\"basic\"
         }
        "
    }
    """

    availableParameters = {"mass","radius","charge"}
    requiredParameters  = set()

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = self.availableParameters,
                         requiredParameters  = self.requiredParameters,
                         **params)

        ############################################################
        ############################################################
        ############################################################

        self.setTypesName("Basic")

        self.addTypesComponent("mass", 1.0)
        self.addTypesComponent("radius", 1.0)
        self.addTypesComponent("charge", 0.0)
